import datetime as _datetime
import linecache
import sys
import typing as t
from dataclasses import dataclass
from re import search
from types import FrameType

import dill
import msgspec
import yaml
from aiopath import AsyncPath

__all__: list[str] = ["load", "dump", "encode", "decode"]


def yaml_encode(
    obj: t.Any,
    *,
    enc_hook: t.Callable[[t.Any], t.Any] | None = None,
    sort_keys: bool = False,
) -> bytes:
    dumper = getattr(yaml, "CSafeDumper", yaml.SafeDumper)
    dumper.add_representer(
        type(None),
        lambda dumper, value: dumper.represent_scalar("tag:yaml.org,2002:null", ""),
    )
    return yaml.dump_all(
        [
            msgspec.yaml._to_builtins(
                obj,
                builtin_types=(_datetime.datetime, _datetime.date),
                enc_hook=enc_hook,
            )
        ],
        encoding="utf-8",
        Dumper=dumper,  # type: ignore
        allow_unicode=True,
        sort_keys=sort_keys,
    )


msgspec.yaml.encode = yaml_encode


@dataclass
class Serializers:
    json: t.Any = msgspec.json
    yaml: t.Any = msgspec.yaml
    msgpack: t.Any = msgspec.msgpack
    pickle: t.Any = dill
    toml: t.Any = msgspec.toml

    def __post_init__(self) -> None:
        self.pickle.encode = dill.dumps
        self.pickle.decode = dill.loads


serializers = Serializers()


class Encode:
    json: t.Callable[..., t.Any]
    yaml: t.Callable[..., t.Any]
    msgpack: t.Callable[..., t.Any]
    toml: t.Callable[..., t.Any]
    pickle: t.Callable[..., t.Any]
    serializers: dict[str, t.Callable[..., t.Any]]
    path: AsyncPath | None = None
    action: str | None = None
    sort_keys: bool = True
    use_list: bool = False
    secure: bool = False
    secret_key: str | None = None
    secure_salt: str | None = None
    serializer: t.Callable[..., t.Any] | None = None

    def __init__(self) -> None:
        self.serializers = serializers.__dict__
        for s in self.serializers:
            setattr(self, s, self.__call__)

    async def process(self, obj: t.Any, **kwargs: object) -> t.Any:
        if self.action in ("load", "decode"):
            if self.serializer is msgspec.msgpack:  # type: ignore
                kwargs["use_list"] = self.use_list
            if isinstance(obj, AsyncPath):
                obj = await obj.read_text()
            return self.serializer.decode(obj, **kwargs)  # type: ignore
        elif self.action in ("dump", "encode"):
            if self.serializer is msgspec.yaml:  # type: ignore
                kwargs["sort_keys"] = self.sort_keys
            data: bytes = self.serializer.encode(obj, **kwargs)  # type: ignore
            if isinstance(self.path, AsyncPath):
                return await self.path.write_bytes(data)
            return data

    def get_vars(self, frame: FrameType) -> tuple[str, t.Any]:
        code_context = linecache.getline(frame.f_code.co_filename, frame.f_lineno)
        pattern = r"await\s(\w+)\.(\w+)\("
        calling_method = search(pattern, code_context)
        return calling_method.group(1), self.serializers[calling_method.group(2)]

    async def __call__(
        self,
        obj: t.Any,
        path: AsyncPath | None = None,
        sort_keys: bool = False,
        use_list: bool = False,
        **kwargs: object,
    ) -> t.Any:
        self.path = path
        self.sort_keys = sort_keys
        self.use_list = use_list
        self.action, self.serializer = self.get_vars(sys._getframe(1))
        return await self.process(obj, **kwargs)  # type: ignore


dump = load = encode = decode = Encode()



import asyncio
import typing as t

from aioconsole import aprint
from rich import box
from rich.console import Console, RenderableType
from rich.padding import Padding
from rich.table import Table
from rich.traceback import install
from acb import pkg_registry
from acb.actions import action_registry
from acb.adapters import get_adapters
from acb.config import Config
from acb.depends import depends


class RichConsole(Console):
    config: Config = depends()

    def __init__(self) -> None:
        super().__init__()
        if not self.config.deployed:
            install(console=self)

    def _write_buffer(self) -> None:
        with self._lock:
            if self.record and not self._buffer_index:
                with self._record_buffer_lock:
                    self._record_buffer.extend(self._buffer[:])

            if self._buffer_index == 0:
                text = self._render_buffer(self._buffer[:])
                try:
                    asyncio.run(aprint(text))
                except UnicodeEncodeError as error:
                    error.reason = (
                        f"{error.reason}\n*** You may need to add"
                        f" PYTHONIOENCODING=utf-8 to your environment ***"
                    )
                    raise
                self.file.flush()
                del self._buffer[:]


console = RichConsole()

table_args: dict[str, t.Any] = {
    "show_lines": True,
    "box": box.ROUNDED,
    "min_width": 100,
    "border_style": "bold blue",
}


@depends.inject
def display_components(config: Config = depends()) -> None:
    if not config.deployed:
        pkgs = Table(
            title="[b][u bright_white]A[/u bright_white][bright_green]synchronous "
            "[u bright_white]C[/u bright_white][bright_green]omponent "
            "[u bright_white]B[/u bright_white][bright_green]ase[/b]",
            **table_args,
        )
        for prop in ("Pkg", "Path"):
            pkgs.add_column(prop)
        for i, pkg in enumerate(pkg_registry.get()):
            pkgs.add_row(
                pkg.name,
                str(pkg.path),
                style="bold white" if i % 2 else "bold white on blue",
            )
        pkgs_padded: RenderableType = Padding(pkgs, (2, 4, 0, 4))
        console.print(pkgs_padded)

        actns = Table(
            title="[b][u bright_white]A[/u bright_white][bright_green]ctions",
            **table_args,
        )
        for prop in ("Name", "Pkg", "Methods"):
            actns.add_column(prop)
        for i, action in enumerate(action_registry.get()):
            actns.add_row(
                action.name,
                action.pkg,
                ", ".join(action.methods),
                style="bold white" if i % 2 else "bold white on blue",
            )
        actns_padded: RenderableType = Padding(actns, (0, 4, 1, 4))
        console.print(actns_padded)

        adptrs = Table(
            title="[b][u bright_white]A[/u bright_white][bright_green]datpters",
            **table_args,
        )
        for prop in ("Category", "Name", "Pkg"):
            adptrs.add_column(prop)
        for i, adapter in enumerate(get_adapters()):
            adptrs.add_row(
                adapter.category,
                adapter.name,
                adapter.pkg,
                style="bold white" if i % 2 else "bold white on blue",
            )
        adptrs_padded: RenderableType = Padding(adptrs, (0, 4, 1, 4))
        console.print(adptrs_padded)



import typing as t
from unittest.mock import MagicMock, Mock, patch

import pytest
from rich.console import Console
from acb.console import RichConsole


@pytest.fixture
def testable_rich_console():
    class TestableRichConsole(Console):
        def __init__(
            self,
            test_buffer: t.Any,
            buffer_index: int,
            record_mode: bool,
        ) -> None:
            super().__init__()
            self.test_buffer = test_buffer
            self.test_buffer_index = buffer_index
            self.record = record_mode
            self.test_record_buffer = []
            self._lock = MagicMock()
            self._record_buffer_lock = MagicMock()
            self._lock.__enter__ = MagicMock(return_value=None)
            self._lock.__exit__ = MagicMock(return_value=None)
            self._record_buffer_lock.__enter__ = MagicMock(return_value=None)
            self._record_buffer_lock.__exit__ = MagicMock(return_value=None)
            self.file = MagicMock()
            self.file.flush = MagicMock()

        def _render_buffer(self, buffer: t.Any) -> str:
            return "rendered content"

        def _write_buffer(self) -> None:
            with self._lock:
                if self.record and not self.test_buffer_index:
                    with self._record_buffer_lock:
                        self.test_record_buffer.extend(self.test_buffer[:])

                if self.test_buffer_index == 0:
                    text = self._render_buffer(self.test_buffer[:])
                    try:
                        print(text)
                    except UnicodeEncodeError as error:
                        error.reason = (
                            f"{error.reason}\n*** You may need to add"
                            f" PYTHONIOENCODING=utf-8 to your environment ***"
                        )
                        raise
                    self.file.flush()
                    del self.test_buffer[:]

    return TestableRichConsole


class TestRichConsole:
    def test_init_with_deployed_config(self) -> None:
        mock_config = Mock()
        mock_config.deployed = True

        with patch("acb.console.depends") as mock_depends:
            mock_depends.return_value = mock_config
            with patch("acb.console.install") as mock_install:
                with patch.object(
                    RichConsole, "__init__", return_value=None
                ) as mock_init:
                    rich_console = RichConsole()

                    mock_init.side_effect = None
                    RichConsole.__init__(rich_console)

                mock_install.assert_not_called()
                assert isinstance(rich_console, Console)

    def test_init_with_non_deployed_config(self) -> None:
        mock_config = Mock()
        mock_config.deployed = False

        with patch("acb.console.depends") as mock_depends:
            mock_depends.return_value = mock_config
            with patch("acb.console.install") as mock_install:
                rich_console = RichConsole()

                mock_install.assert_called_once_with(console=rich_console)
                assert isinstance(rich_console, Console)

    def test_init_with_string_config(self) -> None:
        mock_config = "string_config"

        with patch("acb.console.depends") as mock_depends:
            mock_depends.return_value = mock_config
            with patch("acb.console.install") as mock_install:
                with patch.object(
                    RichConsole, "__init__", return_value=None
                ) as mock_init:
                    rich_console = RichConsole()

                    mock_init.side_effect = None
                    RichConsole.__init__(rich_console)

                mock_install.assert_not_called()
                assert isinstance(rich_console, Console)

    def test_write_buffer_functionality(self, testable_rich_console: t.Any) -> None:
        test_buffer = ["test buffer content"]

        with patch("acb.console.depends") as mock_depends:
            mock_config = Mock()
            mock_depends.return_value = mock_config

            console = testable_rich_console(test_buffer.copy(), 0, False)

            with patch("builtins.print") as mock_print:
                console._write_buffer()

                mock_print.assert_called_once_with("rendered content")

    def test_write_buffer_with_record(self, testable_rich_console: t.Any) -> None:
        test_buffer = ["test buffer content"]

        with patch("acb.console.depends") as mock_depends:
            mock_config = Mock()
            mock_depends.return_value = mock_config

            console = testable_rich_console(test_buffer.copy(), 0, True)

            with patch("builtins.print") as mock_print:
                console._write_buffer()

                mock_print.assert_called_once_with("rendered content")
                assert console.test_record_buffer

    def test_write_buffer_with_unicode_error(
        self, testable_rich_console: t.Any
    ) -> None:
        test_buffer = ["test buffer content"]

        with patch("acb.console.depends") as mock_depends:
            mock_config = Mock()
            mock_depends.return_value = mock_config

            console = testable_rich_console(test_buffer.copy(), 0, False)

            unicode_error = UnicodeEncodeError("utf-8", "test", 0, 1, "test error")
            mock_print = Mock(side_effect=unicode_error)

            with (
                patch("builtins.print", mock_print),
                pytest.raises(UnicodeEncodeError) as exc_info,
            ):
                console._write_buffer()

            mock_print.assert_called_once_with("rendered content")
            assert "PYTHONIOENCODING=utf-8" in exc_info.value.reason



import os
from contextvars import ContextVar
from inspect import currentframe

from aiopath import AsyncPath
from pydantic import BaseModel
from acb.actions import Action, register_actions
from acb.adapters import Adapter, register_adapters

_deployed: bool = os.getenv("DEPLOYED", "False").lower() == "true"
_testing: bool = os.getenv("TESTING", "False").lower() == "true"


class Pkg(BaseModel, arbitrary_types_allowed=True):
    name: str
    path: AsyncPath
    actions: list[Action] = []
    adapters: list[Adapter] = []


pkg_registry: ContextVar[list[Pkg]] = ContextVar("pkg_registry", default=[])


def register_pkg() -> None:
    path = AsyncPath(currentframe().f_back.f_code.co_filename).parent
    name = path.stem
    registry = pkg_registry.get()
    if name not in [p.name for p in registry]:
        actions = register_actions()
        adapters = register_adapters()
        pkg = Pkg(name=name, path=path, actions=actions, adapters=adapters)
        if pkg.name == "acb":
            registry.append(pkg)
        else:
            registry.insert(1, pkg)



from unittest.mock import AsyncMock, patch

import pytest
from aiopath import AsyncPath
from pydantic import SecretStr
from acb.config import (
    AppSettings,
    Config,
    FileSecretSource,
    Platform,
    Settings,
    YamlSettingsSource,
    gen_password,
    get_version,
)


@pytest.mark.asyncio
async def test_get_version_with_pyproject(monkeypatch: pytest.MonkeyPatch) -> None:
    AsyncPath("/fake/path/pyproject.toml")
    AsyncPath("/fake/path/_version")

    with (
        patch("acb.config.load.toml", new_callable=AsyncMock) as mock_load_toml,
        patch.object(AsyncPath, "exists", new_callable=AsyncMock) as mock_exists,
        patch.object(AsyncPath, "write_text", new_callable=AsyncMock) as mock_write,
    ):
        mock_exists.return_value = True
        mock_load_toml.return_value = {"project": {"version": "1.2.3"}}

        result = await get_version()

        monkeypatch.setattr("acb.config.root_path", AsyncPath("/fake/path"))

        assert result == "1.2.3"
        mock_load_toml.assert_called_once()
        mock_write.assert_called_once_with("1.2.3")


@pytest.mark.asyncio
async def test_get_version_with_version_file() -> None:
    with (
        patch("acb.config.root_path", AsyncPath("/fake/path")),
        patch.object(AsyncPath, "exists", new_callable=AsyncMock) as mock_exists,
        patch.object(AsyncPath, "read_text", new_callable=AsyncMock) as mock_read,
    ):
        mock_exists.side_effect = [False, True]
        mock_read.return_value = "2.3.4"

        result = await get_version()

        assert result == "2.3.4"
        mock_read.assert_called_once()


@pytest.mark.asyncio
async def test_get_version_fallback() -> None:
    with (
        patch("acb.config.root_path", AsyncPath("/fake/path")),
        patch.object(AsyncPath, "exists", new_callable=AsyncMock) as mock_exists,
    ):
        mock_exists.return_value = False

        result = await get_version()

        assert result == "0.0.1"


def test_platform_enum() -> None:
    assert Platform.aws.value == "aws"
    assert Platform.gcp.value == "gcp"
    assert Platform.azure.value == "azure"

    aws_value = Platform.aws
    gcp_value = Platform.gcp
    azure_value = Platform.azure

    assert aws_value == "aws"
    assert gcp_value == "gcp"
    assert azure_value == "azure"


def test_gen_password() -> None:
    password = gen_password()
    assert isinstance(password, str)
    assert len(password) >= 10

    custom_size = 20
    password = gen_password(custom_size)
    assert isinstance(password, str)
    assert len(password) >= custom_size


@pytest.mark.asyncio
async def test_settings_initialization() -> None:
    with patch(
        "acb.config.Settings._settings_build_values", new_callable=AsyncMock
    ) as mock_build:
        mock_build.return_value = {"test_key": "test_value"}

        settings = Settings()

        mock_build.assert_called_once()
        assert settings.test_key == "test_value"


@pytest.mark.asyncio
async def test_app_settings_validation_valid_name() -> None:
    app_settings = AppSettings(name="valid-app-name")
    assert app_settings.name == "valid-app-name"
    assert app_settings.title == "Valid App Name"


@pytest.mark.asyncio
async def test_app_settings_validation_invalid_name_too_short() -> None:
    with pytest.raises(SystemExit, match="App name to short"):
        AppSettings(name="ab")


@pytest.mark.asyncio
async def test_app_settings_validation_invalid_name_too_long() -> None:
    long_name = "a" * 64
    with pytest.raises(SystemExit, match="App name to long"):
        AppSettings(name=long_name)


@pytest.mark.asyncio
async def test_app_settings_validation_name_with_special_chars() -> None:
    with (
        patch.object(
            AppSettings, "_settings_build_values", new_callable=AsyncMock
        ) as mock_build,
        patch.object(
            AppSettings, "cloud_compliant_app_name", return_value="my-app-name123"
        ),
    ):
        mock_build.return_value = {"name": "my-app-name123"}

        app_settings = AppSettings(name="My App_Name.123!")
        assert app_settings.name == "my-app-name123"


async def test_file_secret_source() -> None:
    class MockSettings(Settings):
        mock_password: SecretStr = SecretStr("default_password")
        mock_api_key: SecretStr = SecretStr("default_api_key")
        normal_field: str = "normal"

    with patch("acb.config._testing", True):
        test_secrets_path = AsyncPath("/tmp/test_secrets")  # nosec B108
        file_source = FileSecretSource(MockSettings, secrets_path=test_secrets_path)

        model_secrets = file_source.get_model_secrets()
        assert "mock_mock_password" in model_secrets
        assert "mock_mock_api_key" in model_secrets
        assert "normal_field" not in model_secrets

        result = await file_source()

        assert isinstance(result.get("mock_password"), SecretStr)
        assert result["mock_password"].get_secret_value() == "default_password"

        assert isinstance(result.get("mock_api_key"), SecretStr)
        assert result["mock_api_key"].get_secret_value() == "default_api_key"


async def test_yaml_settings_source_testing_mode() -> None:
    class MockSettings(Settings):
        field1: str = "default1"
        field2: int = 42
        secret_field: SecretStr = SecretStr("secret")

    with (
        patch("acb.config._testing", True),
        patch("acb.config.debug", {}),
        patch("acb.config.project", ""),
        patch("acb.config.app_name", ""),
    ):
        yaml_source = YamlSettingsSource(MockSettings)

        result = await yaml_source.load_yml_settings()

        assert result["field1"] == "default1"
        assert result["field2"] == 42
        assert "secret_field" not in result

        call_result = await yaml_source()
        assert call_result["field1"] == "default1"
        assert call_result["field2"] == 42


@pytest.mark.asyncio
async def test_config_initialization() -> None:
    with (
        patch("acb.config.DebugSettings", return_value="debug_settings"),
        patch("acb.config.AppSettings", return_value="app_settings"),
    ):
        config = Config()
        config.init()

        assert config.debug == "debug_settings"
        assert config.app == "app_settings"



import logging
import typing as t
from inspect import currentframe

from aioconsole import aprint
from loguru._logger import Core as _Core
from loguru._logger import Logger as _Logger
from acb.config import Config, debug, root_path
from acb.depends import depends

from ._base import LoggerBase, LoggerBaseSettings


class LoggerSettings(LoggerBaseSettings):
    serialize: t.Optional[bool] = False
    format: t.Optional[dict[str, str]] = dict(
        time="<b><e>[</e> <w>{time:YYYY-MM-DD HH:mm:ss.SSS}</w> <e>]</e></b>",
        level=" <level>{level:>8}</level>",
        sep=" <b><w>in</w></b> ",
        name="<b>{extra[mod_name]:>20}</b>",
        line="<b><e>[</e><w>{line:^5}</w><e>]</e></b>",
        message="  <level>{message}</level>",
    )
    level_per_module: t.Optional[dict[str, str | None]] = {}
    level_colors: t.Optional[dict[str, str]] = {}
    settings: t.Optional[dict[str, t.Any]] = {}

    @depends.inject
    def __init__(self, config: Config = depends(), **values: t.Any) -> None:
        super().__init__(**values)
        self.log_level = (
            self.deployed_level.upper()
            if config.deployed or config.debug.production
            else self.log_level
        )
        self.level_per_module = {
            m: "DEBUG" if v else self.log_level for m, v in debug.items()
        }
        self.settings = dict(
            format="".join(self.format.values()),
            enqueue=True,
            backtrace=False,
            catch=False,
            serialize=self.serialize,
            diagnose=False,
            colorize=True,
        )


class Logger(_Logger, LoggerBase):
    def __init__(self) -> None:
        super().__init__(
            core=_Core(),
            exception=None,
            depth=0,
            record=False,
            lazy=False,
            colors=False,
            raw=False,
            capture=True,
            patchers=[],
            extra={},
        )

    @staticmethod
    async def async_sink(message: str) -> None:
        await aprint(message, end="")

    @depends.inject
    async def init(self, config: Config = depends()) -> None:
        def patch_name(record: dict[str, t.Any]) -> str:  # type: ignore
            mod_parts = record["name"].split(".")
            mod_name = ".".join(mod_parts[:-1])
            if len(mod_parts) > 3:
                mod_name = ".".join(mod_parts[1:-1])
            return mod_name.replace("_sdk", "")

        def filter_by_module(record: dict[str, t.Any]) -> bool:
            try:
                name = record["name"].split(".")[-2]
            except IndexError:
                name = record["name"]
            level_ = config.logger.log_level
            if name in config.logger.level_per_module:
                level_ = config.logger.level_per_module[name]
            try:
                levelno_ = self.level(level_).no
            except ValueError:
                raise ValueError(
                    f"The filter dict contains a module '{name}' associated to a level "
                    f" name which does not exist: '{level_}'"
                )
            if level_ is False:
                return False
            return record["level"].no >= levelno_

        self.remove()
        self.configure(
            patcher=lambda record: record["extra"].update(mod_name=patch_name(record)),
        )
        self.add(
            self.async_sink,
            filter=filter_by_module,  # type: ignore
            **config.logger.settings,
        )
        for level, color in config.logger.level_colors.items():
            self.level(level.upper(), color=f"[{color}]")
        if config.debug.logger:
            self.debug("debug")
            self.info("info")
            self.warning("warning")
            self.error("error")
            self.critical("critical")
        self.info(f"App path: {root_path}")
        self.info(f"App deployed: {config.deployed}")


depends.set(Logger)


class InterceptHandler(logging.Handler):
    def emit(self, record: logging.LogRecord) -> None:
        logger = depends.get(Logger)
        try:
            level = logger.level(record.levelname).name
        except ValueError:
            level = record.levelno

        frame, depth = currentframe(), 0
        while frame and (depth == 0 or frame.f_code.co_filename == logging.__file__):
            frame = frame.f_back
            depth += 1

        logger.opt(depth=depth, exception=record.exc_info).log(
            level, record.getMessage()
        )


logging.basicConfig(handlers=[InterceptHandler()], level=0, force=True)



from pydantic import field_validator

from ._base import NosqlBase, NosqlBaseSettings


class NosqlSettings(NosqlBaseSettings):
    cache_db: int = 0

    @field_validator("cache_db")
    @classmethod
    def cache_db_not_zero(cls, v: int) -> int:  # noqa: F841
        if v < 3 and v != 0:
            raise ValueError("must be > 3 (0-2 are reserved)")
        return 0


class Nosql(NosqlBase):
    def __init__(self) -> None: ...



import typing as t

from aiocache.backends.redis import RedisBackend
from aiocache.serializers import PickleSerializer
from coredis.cache import TrackingCache
from coredis.client import Redis, RedisCluster
from pydantic import SecretStr
from acb.config import Config
from acb.debug import debug
from acb.depends import depends

from ._base import CacheBase, CacheBaseSettings


class CacheSettings(CacheBaseSettings):
    host: SecretStr = SecretStr("127.0.0.1")
    local_host: str = "127.0.0.1"
    port: t.Optional[int] = 6379
    cluster: t.Optional[bool] = False
    connect_timeout: t.Optional[float] = None
    max_connections: t.Optional[int] = None
    health_check_interval: t.Optional[int] = 0

    @depends.inject
    def __init__(self, config: Config = depends(), **values: t.Any) -> None:
        super().__init__(**values)
        self.host = SecretStr(self.local_host) if not config.deployed else self.host


class Cache(CacheBase, RedisBackend):  # type: ignore
    async def _close(self, *args: t.Any, _conn: t.Any = None, **kwargs: t.Any) -> None:
        pass

    async def _clear(
        self, namespace: t.Optional[str] = None, _conn: t.Any = None
    ) -> t.Literal[True]:
        if not namespace:
            pattern = f"{self.config.app.name}:*"
        else:
            pattern = f"{self.config.app.name}:{namespace}:*"
        keys = await self.client.keys(pattern)
        if keys:
            debug(keys)
            for key in keys:
                await self.client.unlink((key,))
        return True

    async def _exists(self, key: str, _conn: t.Any = None) -> bool:
        number = await self.client.exists([key])
        return bool(number)

    async def init(self, *args: t.Any, **kwargs: t.Any) -> None:
        super().__init__(
            serializer=PickleSerializer(),
            namespace=f"{self.config.app.name}:",
            **kwargs,
        )
        redis_kwargs = dict(
            host=self.config.cache.host.get_secret_value(),
            port=self.config.cache.port,
            client_name=self.config.app.name,
            cache=TrackingCache(),
            decode_responses=False,
            connect_timeout=self.config.cache.connect_timeout,
            max_connections=self.config.cache.max_connections,
            health_check_interval=self.config.cache.health_check_interval,
        )
        if self.config.cache.cluster:
            self.logger.info("RedisCluster mode enabled")
            del redis_kwargs["health_check_interval"]
            self.client = RedisCluster(**redis_kwargs)  # type: ignore
        else:
            self.client = Redis(**redis_kwargs)  # type: ignore


depends.set(Cache)



import asyncio
import typing as t
from pathlib import Path
from time import sleep
from unittest.mock import AsyncMock, MagicMock, Mock, patch

import pytest
from acb.debug import (
    colorized_stderr_print,
    debug,
    get_calling_module,
    patch_record,
    print_debug_info,
    timeit,
)


class TestGetCallingModule:
    def test_get_calling_module_success(self, mock_config: MagicMock) -> None:
        mock_frame = MagicMock()
        mock_frame.f_back.f_back.f_back.f_code.co_filename = "/path/to/module/file.py"

        setattr(mock_config.debug, "module", True)

        with (
            patch("acb.debug.config", mock_config),
            patch("acb.debug.logging.currentframe", return_value=mock_frame),
        ):
            result = get_calling_module()

            assert result == Path("/path/to/module")

    def test_get_calling_module_no_debug_config(self, mock_config: MagicMock) -> None:
        mock_frame = MagicMock()
        mock_frame.f_back.f_back.f_back.f_code.co_filename = "/path/to/module/file.py"

        with (
            patch("acb.debug.config", mock_config),
            patch("acb.debug.logging.currentframe", return_value=mock_frame),
            patch("acb.debug.getattr", return_value=None),
        ):
            result = get_calling_module()

            assert result is None

    def test_get_calling_module_exception(self, mock_config: MagicMock) -> None:
        with (
            patch("acb.debug.config", mock_config),
            patch("acb.debug.logging.currentframe", side_effect=AttributeError),
        ):
            result = get_calling_module()

            assert result is None


class TestPatchRecord:
    def test_patch_record_with_loguru(
        self, mock_logger: MagicMock, mock_adapter_registry: t.Any
    ) -> None:
        mock_mod = MagicMock(spec=Path)
        mock_mod.name = "module"
        msg = "Test debug message"

        mock_logger.patch.reset_mock()
        mock_logger.debug.reset_mock()

        patched_logger = mock_logger
        mock_logger.patch.return_value = patched_logger

        with patch("acb.debug.adapter_registry") as mock_registry:
            mock_registry.get.return_value = mock_adapter_registry

            with patch("acb.debug.any", return_value=True) as mock_any:
                patch_record(mock_mod, msg, logger=mock_logger)

                assert mock_logger.patch.call_count == 1
                patched_logger.debug.assert_called_once_with(msg)
                mock_any.assert_called_once()

    def test_patch_record_without_loguru(
        self, mock_logger: MagicMock, mock_adapter_registry: t.Any
    ) -> None:
        mock_mod = MagicMock(spec=Path)
        mock_mod.name = "module"
        msg = "Test debug message"

        mock_logger.patch.reset_mock()

        with patch("acb.debug.adapter_registry") as mock_registry:
            mock_registry.get.return_value = mock_adapter_registry

            with patch("acb.debug.any", return_value=False) as mock_any:
                patch_record(mock_mod, msg, logger=mock_logger)

                mock_logger.patch.assert_not_called()
                mock_any.assert_called_once()

    def test_patch_record_exception(
        self, mock_logger: MagicMock, mock_adapter_registry: t.Any
    ) -> None:
        mock_mod = MagicMock(spec=Path)
        mock_mod.name = "module"
        msg = "Test debug message"

        mock_logger.patch.reset_mock()

        mock_logger.patch.side_effect = Exception("Test exception")

        with patch("acb.debug.adapter_registry") as mock_registry:
            mock_registry.get.return_value = mock_adapter_registry

            with patch("acb.debug.any", return_value=True) as mock_any:
                patch_record(mock_mod, msg, logger=mock_logger)

                assert mock_logger.patch.call_count == 1
                mock_any.assert_called_once()


class TestColorizedStderrPrint:
    def test_colorized_stderr_print(self) -> None:
        test_string = "Test colored string"

        async def mock_aprint_coro():
            return None

        mock_aprint = AsyncMock(side_effect=mock_aprint_coro)

        with (
            patch("acb.debug.supportTerminalColorsInWindows") as mock_support_colors,
            patch(
                "acb.debug.colorize", return_value=f"COLORED({test_string})"
            ) as mock_colorize,
            patch("acb.debug.aprint", mock_aprint),
            patch("acb.debug.asyncio.run") as mock_run,
        ):
            colorized_stderr_print(test_string)

            mock_support_colors.assert_called_once()
            mock_support_colors().__enter__.assert_called_once()
            mock_support_colors().__exit__.assert_called_once()

            mock_colorize.assert_called_once_with(test_string)

            mock_aprint.assert_called_once_with(
                f"COLORED({test_string})", use_stderr=True
            )

            assert mock_run.call_count == 1
            args, _ = mock_run.call_args
            assert asyncio.iscoroutine(args[0])


class TestPrintDebugInfo:
    def test_print_debug_info_production(self, mock_config: MagicMock) -> None:
        msg = "Test debug message"
        mock_mod = Path("/path/to/module")
        mock_config.deployed = True

        mock_patch_record = Mock()

        with (
            patch("acb.debug.config", mock_config),
            patch("acb.debug.get_calling_module", return_value=mock_mod),
            patch("acb.debug.patch_record", mock_patch_record),
            patch("acb.debug.colorized_stderr_print") as mock_colorized_print,
        ):
            print_debug_info(msg)

            mock_patch_record.assert_called_once_with(mock_mod, msg)
            mock_colorized_print.assert_not_called()

    def test_print_debug_info_development(self, mock_config: MagicMock) -> None:
        msg = "Test debug message"
        mock_mod = Path("/path/to/module")
        mock_config.deployed = False
        mock_config.debug.production = False

        mock_colorized_print = Mock()

        with (
            patch("acb.debug.config", mock_config),
            patch("acb.debug.get_calling_module", return_value=mock_mod),
            patch("acb.debug.patch_record") as mock_patch_record,
            patch("acb.debug.colorized_stderr_print", mock_colorized_print),
        ):
            print_debug_info(msg)

            mock_patch_record.assert_not_called()
            mock_colorized_print.assert_called_once_with(msg)

    def test_print_debug_info_no_module(self, mock_config: MagicMock) -> None:
        msg = "Test debug message"

        mock_patch_record = Mock()
        mock_colorized_print = AsyncMock()

        with (
            patch("acb.debug.config", mock_config),
            patch("acb.debug.get_calling_module", return_value=None),
            patch("acb.debug.patch_record", mock_patch_record),
            patch("acb.debug.colorized_stderr_print", mock_colorized_print),
        ):
            result = print_debug_info(msg)

            mock_patch_record.assert_not_called()
            mock_colorized_print.assert_not_awaited()
            assert result is None


class TestTimeit:
    def test_timeit_decorator(self, mock_logger: MagicMock) -> None:
        def test_function() -> str:
            sleep(0.01)
            return "test result"

        mock_logger.debug.reset_mock()

        decorated_function = timeit(test_function, logger=mock_logger)

        result = decorated_function()

        assert result == "test result"

        assert mock_logger.debug.call_count >= 1

        debug_messages = [call[0][0] for call in mock_logger.debug.call_args_list]
        assert any("test_function" in msg for msg in debug_messages)
        assert any("executed in" in msg for msg in debug_messages)
        assert any("s" in msg for msg in debug_messages)

    @pytest.mark.asyncio
    async def test_timeit_with_args_and_kwargs(self, mock_logger: MagicMock) -> None:
        def test_function_with_args(
            arg1: str, arg2: str, kwarg1: object = None, kwarg2: object = None
        ) -> str:
            async def _test_function_with_args() -> str:
                sleep(0.01)
                return f"{arg1}-{arg2}-{kwarg1}-{kwarg2}"

            return asyncio.run(_test_function_with_args())

        mock_logger.debug.reset_mock()

        decorated_function = timeit(test_function_with_args, logger=mock_logger)

        result = decorated_function("a", "b", kwarg1="c", kwarg2="d")

        assert result == "a-b-c-d"

        assert mock_logger.debug.call_count >= 1

        debug_messages: t.List[str] = [
            call[0][0] for call in mock_logger.debug.call_args_list
        ]
        assert any("test_function_with_args" in msg for msg in debug_messages)


class TestDebugConfiguration:
    def test_debug_configuration(self, mock_config: MagicMock) -> None:
        mock_config.deployed = False
        mock_config.debug.production = False

        mock_configure = MagicMock()

        with (
            patch("acb.debug.config", mock_config),
            patch("acb.debug.debug.configureOutput", mock_configure),
        ):
            from acb.debug import debug_args

            debug.configureOutput(
                prefix="    debug:  ",
                includeContext=True,
                **debug_args,
            )

            assert mock_configure.call_count >= 1

            mock_config.deployed = True

            if mock_config.deployed or mock_config.debug.production:
                debug.configureOutput(
                    prefix="",
                    includeContext=False,
                    **debug_args,
                )

            assert mock_configure.call_count >= 2



import typing as t
from unittest.mock import Mock, patch

from acb.depends import Depends, DependsProtocol, depends


class TestDependsProtocol:
    def test_protocol_checking(self) -> None:
        assert isinstance(depends, DependsProtocol)

        class NonCompliant:
            pass

        non_compliant = NonCompliant()
        assert not isinstance(non_compliant, DependsProtocol)


class TestDepends:
    def test_inject_decorator(self) -> None:
        mock_inject = Mock(return_value=lambda x: x)

        def test_func() -> str:
            return "test"

        with patch("acb.depends.inject_dependency", mock_inject):
            Depends.inject(test_func)

        mock_inject.assert_called_once_with(test_func)

    def test_set_method(self) -> None:
        mock_repository = Mock()
        mock_repository.set.return_value = "set_result"

        class TestClass:
            pass

        with patch("acb.depends.get_repository", return_value=mock_repository):
            result = Depends.set(TestClass)

        mock_repository.set.assert_called_once()
        assert result == "set_result"

    def test_get_method_with_class(self) -> None:
        mock_repository = Mock()
        mock_repository.get.return_value = "get_result"

        class TestClass:
            pass

        with patch("acb.depends.get_repository", return_value=mock_repository):
            result = Depends.get(TestClass)

        mock_repository.get.assert_called_once_with(TestClass)
        assert result == "get_result"

    def test_get_method_with_string(self) -> None:
        mock_repository = Mock()
        mock_repository.get.return_value = "get_result"
        mock_import_adapter = Mock(return_value="imported_adapter")

        with (
            patch("acb.depends.get_repository", return_value=mock_repository),
            patch("acb.adapters.import_adapter", mock_import_adapter),
        ):
            result = Depends.get("adapter_name")

        mock_import_adapter.assert_called_once_with(["adapter_name"])
        mock_repository.get.assert_called_once_with("imported_adapter")
        assert result == "get_result"

    def test_call_method(self) -> None:
        mock_dependency = Mock(return_value="dependency_result")

        with patch("acb.depends.dependency", mock_dependency):
            result = depends()

        mock_dependency.assert_called_once()
        assert result == "dependency_result"


class TestDependsIntegration:
    def test_inject_decorator_usage(self) -> None:
        class TestDependency:
            def method(self) -> str:
                return "dependency_method_called"

        test_dependency = TestDependency()

        def mock_inject(func: t.Callable[..., t.Any]):
            def wrapper(*args: t.Sequence[t.Any], **kwargs: t.Any):
                kwargs["dep"] = test_dependency
                return func(*args, **kwargs)

            return wrapper

        with patch("acb.depends.inject_dependency", mock_inject):

            @depends.inject
            def test_function(dep: TestDependency | None = None):
                return dep.method()

            result = test_function()

        assert result == "dependency_method_called"



from acb.depends import depends

from .loguru import Logger as LoguruLogger
from .loguru import LoggerSettings as LoguruSettings


class LoggerSettings(LoguruSettings):
    serialize: bool | None = True


class Logger(LoguruLogger): ...


depends.set(Logger)



import typing as t
from inspect import stack

from bevy import dependency, get_repository
from bevy import inject as inject_dependency


@t.runtime_checkable
class DependsProtocol(t.Protocol):
    @staticmethod
    def inject(func: t.Callable[..., t.Any]) -> t.Callable[..., t.Any]: ...

    @staticmethod
    def set(class_: t.Any) -> t.Any: ...

    @staticmethod
    def get(class_: t.Any) -> t.Any: ...

    def __call__(self, *args: t.Any, **kwargs: t.Any) -> t.Any: ...


class Depends:
    @staticmethod
    def inject(func: t.Callable[..., t.Any]) -> t.Callable[..., t.Any]:
        return inject_dependency(func)

    @staticmethod
    def set(class_: t.Any) -> t.Any:
        return get_repository().set(class_, class_())

    @staticmethod
    def get(class_: t.Any = None) -> t.Any:
        classes = []
        if not class_ or isinstance(class_, str):
            _classes = []
            if not class_:
                _classes = [
                    c.strip()
                    for c in (
                        stack()[1][4][0].split("=")[0].strip().lower()  # type: ignore
                    ).split(",")
                ]
                _classes = [c.removeprefix("self.") for c in _classes]
            elif isinstance(class_, str):
                _classes = [class_]
            from acb.adapters import import_adapter

            config = False
            index = 0
            if "config" in _classes:
                index = _classes.index("config")
                del _classes[index]
                config = True
            if _classes:
                _classes = import_adapter(_classes)
                classes = [_classes] if not isinstance(_classes, list) else _classes
            if config:
                from acb.config import Config

                classes.insert(index, Config)  # type: ignore
        else:
            classes = [class_]
        classes = [t.cast(c, get_repository().get(c)) for c in classes]
        return classes[0] if len(classes) < 2 else classes

    def __call__(self, *args: t.Any, **kwargs: t.Any) -> t.Any:
        return dependency()


depends = Depends()



import typing as t

from aiopath import AsyncPath
from acb.config import AdapterBase, Settings, app_name


class SecretBaseSettings(Settings):
    secrets_path: AsyncPath


class SecretProtocol(t.Protocol):
    async def list(self, adapter: str) -> t.NoReturn: ...

    async def create(self, name: str, value: str) -> t.NoReturn: ...

    async def update(self, name: str, value: str) -> t.NoReturn: ...

    async def get(self, name: str) -> t.NoReturn: ...

    async def delete(self, name: str) -> t.NoReturn: ...


class SecretBase(AdapterBase):
    prefix: str = f"{app_name}_"



import typing as t
from contextvars import ContextVar
from importlib import import_module
from inspect import isclass

from aiopath import AsyncPath
from sqlmodel import SQLModel
from acb.adapters import get_adapter, get_installed_adapters, import_adapter, root_path
from acb.depends import depends

from ._base import ModelsBase, ModelsBaseSettings

imported_models: ContextVar[list[t.Any]] = ContextVar("imported_models", default=[])

base_models_paths = [AsyncPath(root_path / p) for p in ("models", "models.py")]

if get_adapter("schemas"):
    import_adapter("schemas")


class ModelsSettings(ModelsBaseSettings): ...


class Models(ModelsBase):
    async def import_models(self, path: AsyncPath) -> None:
        depth = -2
        if "adapters" in path.parts:
            depth = -5
        module_path = ".".join(path.parts[depth:]).removesuffix(".py")
        module = import_module(module_path)
        _imported_models = []
        for name, model in {
            n: m
            for n, m in vars(module).items()
            if isclass(m)
            and issubclass(m, SQLModel)
            and hasattr(m, "__table__")
            and n not in imported_models.get()
        }.items():
            setattr(self.sql, name, model)
            _imported_models.append(name)
            self.logger.debug(f"{name} model imported")
        imported_models.get().extend(_imported_models)

    async def init(self) -> None:
        SQLModel.metadata.clear()
        adapter_paths = []
        for adapter in get_installed_adapters():
            model_paths = [
                p
                async for p in adapter.path.parent.iterdir()
                if p.name.startswith("_models")
            ]
            adapter_paths.extend(model_paths)
        for path in [p for p in adapter_paths + base_models_paths if await p.exists()]:
            if await path.is_dir():
                for _path in [
                    s
                    async for s in path.iterdir()
                    if s.suffix == ".py"
                    and await s.is_file()
                    and not s.stem.startswith("_")
                ]:
                    await self.import_models(_path)
            elif await path.is_file():
                await self.import_models(path)
        self.logger.info(f"{len(imported_models.get())} models imported")


depends.set(Models)



import typing as t
from contextlib import asynccontextmanager
from functools import cached_property

from pydantic import SecretStr
from sqlalchemy import log as sqlalchemy_log
from sqlalchemy import pool, text
from sqlalchemy.engine import URL
from sqlalchemy.ext.asyncio import AsyncConnection, AsyncEngine, create_async_engine
from sqlalchemy_utils import create_database, database_exists
from sqlmodel import SQLModel
from sqlmodel.ext.asyncio.session import AsyncSession
from acb.adapters import import_adapter
from acb.config import AdapterBase, Config, Settings, gen_password
from acb.debug import debug
from acb.depends import depends


class SqlBaseSettings(Settings):
    _driver: str
    _async_driver: str
    port: t.Optional[int] = 3306
    pool_pre_ping: t.Optional[bool] = False
    poolclass: t.Optional[t.Any] = None
    host: SecretStr = SecretStr("127.0.0.1")
    local_host: str = "127.0.0.1"
    user: SecretStr = SecretStr("root")
    password: SecretStr = SecretStr(gen_password())
    _url: t.Optional[URL] = None
    _async_url: t.Optional[URL] = None
    engine_kwargs: dict[str, t.Any] = {}
    backup_enabled: bool = False
    backup_bucket: str | None = None
    cloudsql_instance: t.Optional[str] = None
    cloudsql_proxy: t.Optional[bool] = False
    cloudsql_proxy_port: t.Optional[int] = None

    @depends.inject
    def __init__(self, config: Config = depends(), **values: t.Any) -> None:
        super().__init__(**values)
        if self.cloudsql_proxy:
            self.port = self.cloudsql_proxy_port if not config.deployed else self.port
        url_kwargs = dict(
            drivername=self._driver,
            username=self.user.get_secret_value(),
            password=self.password.get_secret_value(),
            host=(
                self.local_host if not config.deployed else self.host.get_secret_value()
            ),
            port=self.port,
            database=config.app.name,
        )
        self.poolclass = getattr(pool, self.poolclass) if self.poolclass else None
        self._url = URL.create(**url_kwargs)  # type: ignore
        async_url_kwargs = dict(drivername=self._async_driver)
        self._async_url = URL.create(**(url_kwargs | async_url_kwargs))  # type: ignore
        self.engine_kwargs["echo"] = (
            "debug" if config.logger.verbose else config.debug.sql
        )
        self.engine_kwargs["echo_pool"] = (
            "debug" if config.logger.verbose else config.debug.sql
        )
        self.engine_kwargs = (  # type: ignore
            dict(poolclass=self.poolclass, pool_pre_ping=self.pool_pre_ping)
            | self.engine_kwargs
        )


class SqlBase(AdapterBase):
    exists: bool = False

    @cached_property
    def engine(self) -> AsyncEngine:
        self.logger.debug(self.config.sql._async_url)
        if not database_exists(self.config.sql._url):
            self.logger.debug(self.config.sql._async_url)
            create_database(self.config.sql._url)
        self.exists = True
        return create_async_engine(
            self.config.sql._async_url, **self.config.sql.engine_kwargs
        )

    @cached_property
    def session(self) -> AsyncSession:
        return AsyncSession(self.engine, expire_on_commit=False)

    @asynccontextmanager
    async def get_session(self) -> t.AsyncGenerator[AsyncSession, None]:
        async with self.session as sess:
            yield sess

    @asynccontextmanager
    async def get_conn(self) -> t.AsyncGenerator[AsyncConnection, None]:
        async with self.engine.begin() as conn:
            yield conn

    async def init(
        self,
    ) -> None:
        sqlalchemy_log._add_default_handler = lambda _: None  # type: ignore
        async with self.get_conn() as conn:
            if self.config.debug.sql:
                ps = await conn.execute(text("SHOW FULL PROCESSLIST"))
                show_ps = [p for p in ps]
                debug(show_ps)
                ids = [
                    a[0]
                    for a in show_ps
                    if (
                        len(a) > 3
                        and a[1] == self.config.sql.user.get_secret_value()
                        and a[3] == self.config.app.name
                    )
                ]
                debug(ids)
            try:
                await conn.run_sync(SQLModel.metadata.drop_all)
                import_adapter("models")
                await conn.run_sync(SQLModel.metadata.create_all)
            except Exception as e:
                self.logger.error(e)
                raise e



import typing as t

from acb.config import Settings


class LoggerBaseSettings(Settings):
    verbose: bool = False
    deployed_level: str = "WARNING"
    log_level: t.Optional[str] = "INFO"


class LoggerBase:
    def debug(self, msg: str, *args: t.Any, **kwargs: t.Any) -> None: ...

    def info(self, msg: str, *args: t.Any, **kwargs: t.Any) -> None: ...

    def warning(self, msg: str, *args: t.Any, **kwargs: t.Any) -> None: ...

    def error(self, msg: str, *args: t.Any, **kwargs: t.Any) -> None: ...

    async def init(self) -> None: ...



import typing as t
from functools import cached_property

import nest_asyncio
from aiopath import AsyncPath
from fsspec.asyn import AsyncFileSystem
from google.cloud.exceptions import NotFound
from acb.actions import hash
from acb.adapters import get_adapter, tmp_path
from acb.config import AdapterBase, Config, Settings, import_adapter
from acb.debug import debug
from acb.depends import depends

Logger = import_adapter()
nest_asyncio.apply()


class StorageBaseSettings(Settings):
    prefix: t.Optional[str] = None
    local_path: t.Optional[AsyncPath] = tmp_path / "storage"
    user_project: t.Optional[str] = None
    buckets: dict[str, str] = {"test": "test-bucket"}
    cors: t.Optional[dict[str, dict[str, list[str] | int]]] = None
    local_fs: t.Optional[bool] = False
    memory_fs: t.Optional[bool] = False

    @depends.inject
    def __init__(self, config: Config = depends(), **values: t.Any) -> None:
        super().__init__(**values)
        self.prefix = self.prefix or config.app.name or ""
        self.user_project = self.user_project or config.app.name or ""
        self.local_fs = get_adapter("storage").name in ("file", "memory")
        self.memory_fs = get_adapter("storage").name == "memory"


class StorageBucket:
    config: Config = depends()
    logger: Logger = depends()

    def __init__(
        self,
        client: t.Any,
        bucket: str,
        prefix: t.Optional[str] = None,
    ) -> None:
        self.client = client
        self.name = bucket
        self.bucket = self.config.storage.buckets[bucket]
        self.prefix = prefix or self.config.storage.prefix
        self.root = AsyncPath(f"{self.bucket}/{self.prefix}")

    def get_name(self, path: AsyncPath) -> str:
        return path.name

    def get_path(self, path: AsyncPath) -> str:
        if self.config.storage.local_fs:
            return str(path)
        return str(self.root / path)

    def get_url(self, path: AsyncPath) -> str:
        return self.client.url(self.get_path(path))

    async def get_date_created(self, path: AsyncPath) -> t.Any:
        return (await self.stat(path))["timeCreated"]

    async def get_date_updated(self, path: AsyncPath) -> t.Any:
        return (await self.stat(path))["updated"]

    async def get_size(self, path: AsyncPath) -> int:
        return (await self.stat(path))["size"]

    @staticmethod
    async def get_checksum(path: AsyncPath) -> int:
        return await hash.crc32c(path)  # type: ignore

    async def get_signed_url(self, path: AsyncPath, expires: int = 3600) -> t.Any:
        return await self.client._sign(self.get_path(path), expires=expires)

    async def stat(self, path: AsyncPath) -> t.Any:
        _path = self.get_path(path)
        if self.config.storage.memory_fs:
            info = self.client.info(_path)
            stat = dict(
                name=info.get("name"),
                size=info.get("size"),
                type=info.get("type"),
                mtime=self.client.modified(_path).timestamp(),
                created=self.client.created(_path).timestamp(),
            )
            return stat
        return await self.client._info(_path)

    async def list(self, dir_path: AsyncPath) -> t.Any:
        return await self.client._ls(self.get_path(dir_path))

    async def exists(self, path: AsyncPath) -> t.Any:
        if get_adapter("storage").name == "memory":
            return self.client.isfile(self.get_path(path))
        return await self.client._exists(self.get_path(path))

    async def create_bucket(self, path: AsyncPath) -> t.Any:
        create_args: dict[str, t.Any] = dict(
            create_parents=True,
            enable_versioning=False,
        )
        if self.name == "media":
            create_args["acl"] = "public-read"
            if self.config.storage.cors:
                self.client.set_cors(self.name, self.config.storage.cors)
        elif self.name == "templates":
            create_args["enable_versioning"] = True
        return await self.client._mkdir(self.get_path(path), **create_args)

    async def open(self, path: AsyncPath) -> t.BinaryIO:
        try:
            async with self.client.open(self.get_path(path), "rb") as f:
                return f.read()
        except (NotFound, FileNotFoundError, RuntimeError):
            raise FileNotFoundError
        except Exception as e:
            debug(e)
            raise e

    async def write(self, path: AsyncPath, data: t.Any) -> t.Any:
        stor_path = self.get_path(path)
        try:
            if get_adapter("storage").name == "memory":
                self.client.pipe_file(stor_path, data)
            else:
                await self.client._pipe_file(stor_path, data)
        except Exception as e:
            debug(e)
            raise e

    async def delete(self, path: AsyncPath) -> t.Any:
        stor_path = self.get_path(path)
        await self.client._rm_file(stor_path)


class StorageProtocol(t.Protocol):
    file_system: t.Any = AsyncFileSystem

    @cached_property
    def client(self) -> AsyncFileSystem: ...
    async def init(self) -> None: ...


class StorageBase(AdapterBase):
    file_system: t.Type[AsyncFileSystem]
    templates: StorageBucket | None = None
    media: StorageBucket | None = None
    test: StorageBucket | None = None

    @cached_property
    def client(self) -> AsyncFileSystem:
        return self.file_system(asynchronous=True)

    async def init(self) -> None:
        for bucket in self.config.storage.buckets:
            setattr(self, bucket, StorageBucket(self.client, bucket))
            self.logger.debug(f"{bucket.title()} storage bucket initialized")


class StorageFile:
    def __init__(self, *, name: str, storage: StorageBucket) -> None:
        self._storage: StorageBucket = storage
        self._name: str = name

    @property
    def name(self) -> str:
        return self._storage.get_name(AsyncPath(self._name))

    @property
    def path(self) -> str:
        return self._storage.get_path(AsyncPath(self._name))

    @property
    async def size(self) -> t.Any:
        return await self._storage.get_size(AsyncPath(self._name))

    @property
    async def checksum(self) -> t.Any:
        return await self._storage.get_checksum(AsyncPath(self._name))

    async def open(self) -> t.Any:
        return await self._storage.open(AsyncPath(self._name))

    async def write(self, file: t.BinaryIO) -> t.Any:
        return await self._storage.write(path=AsyncPath(self._name), data=file)

    def __str__(self) -> str:
        return self.path


class StorageImage(StorageFile):
    def __init__(
        self, *, name: str, storage: StorageBucket, height: int, width: int
    ) -> None:
        super().__init__(name=name, storage=storage)
        self._width = width
        self._height = height

    @property
    def height(self) -> int:
        return self._height

    @property
    def width(self) -> int:
        return self._width



import typing as t

from aiopath import AsyncPath
from pydantic import EmailStr, SecretStr
from acb.config import AdapterBase, Config, Settings, gen_password
from acb.depends import depends


class SmtpBaseSettings(Settings):
    api_key: t.Optional[SecretStr] = None
    mx_servers: t.Optional[list[str]] = []
    password: SecretStr = SecretStr(gen_password())
    domain: t.Optional[str] = None
    port: int = 587
    api_url: t.Optional[str] = ""
    default_from: t.Optional[EmailStr] = None
    default_from_name: t.Optional[str] = None
    test_receiver: t.Optional[EmailStr] = None
    tls: bool = True
    ssl: bool = False
    template_folder: t.Optional[AsyncPath] = None
    forwards: dict[str, EmailStr] = dict(
        admin="pat@example.com",
        info="terry@example.com",
    )

    @depends.inject
    def __init__(self, config: Config = depends(), **values: t.Any) -> None:
        super().__init__(**values)
        self.domain = f"mail.{config.app.domain}"
        self.default_from = f"info@{config.app.domain}"
        self.default_from_name = config.app.title


class SmtpBase(AdapterBase): ...



import typing as t
from pathlib import Path
from unittest.mock import Mock, patch

import pytest
from aiopath import AsyncPath
from acb.adapters import (
    Adapter,
    AdapterNotFound,
    adapter_registry,
    get_adapters,
    get_installed_adapter,
    get_installed_adapters,
    path_adapters,
)


class TestAdapterClass:
    def test_adapter_equality(self) -> None:
        adapter1 = Adapter(
            name="test",
            class_name="Test",
            category="test_category",
            pkg="acb",
            module="acb.adapters.test_category.test",
            path=AsyncPath(Path(__file__)),
        )

        adapter2 = Adapter(
            name="test",
            class_name="Test",
            category="test_category",
            pkg="acb",
            module="acb.adapters.test_category.test",
            path=AsyncPath(Path(__file__)),
        )

        adapter3 = Adapter(
            name="different",
            class_name="Different",
            category="different_category",
            pkg="acb",
            module="acb.adapters.different_category.different",
            path=AsyncPath(Path(__file__)),
        )

        assert adapter1 == adapter2
        assert adapter1 != adapter3
        assert adapter1 != "not_an_adapter"

        assert hash(adapter1) == hash(adapter2)
        assert hash(adapter1) != hash(adapter3)

        assert str(adapter1) == adapter1.__repr__()


class TestAdapterRegistry:
    def setup_method(self) -> None:
        adapter_registry.set([])

    def test_get_adapter(self) -> None:
        adapter1 = Adapter(
            name="test1",
            class_name="Test1",
            category="category1",
            enabled=True,
        )
        adapter2 = Adapter(
            name="test2",
            class_name="Test2",
            category="category2",
            enabled=False,
        )
        adapter_registry.set([adapter1, adapter2])

        with patch("acb.adapters.get_adapter") as mock_get_adapter:

            def side_effect(category: str) -> t.Optional[Adapter]:
                if category == "category1":
                    return adapter1
                elif category == "category2":
                    return None
                return None

            mock_get_adapter.side_effect = side_effect

            from acb.adapters import get_adapter as patched_get_adapter

            assert patched_get_adapter("category1") == adapter1
            assert patched_get_adapter("category2") is None
            assert patched_get_adapter("non_existent") is None

    def test_get_adapters(self) -> None:
        adapter1 = Adapter(
            name="test1",
            class_name="Test1",
            category="category1",
            enabled=True,
        )
        adapter2 = Adapter(
            name="test2",
            class_name="Test2",
            category="category2",
            enabled=False,
        )
        adapter3 = Adapter(
            name="test3",
            class_name="Test3",
            category="category3",
            enabled=True,
        )
        adapter_registry.set([adapter1, adapter2, adapter3])

        enabled_adapters = get_adapters()
        assert len(enabled_adapters) == 2
        assert adapter1 in enabled_adapters
        assert adapter2 not in enabled_adapters
        assert adapter3 in enabled_adapters

    def test_get_installed_adapter(self) -> None:
        adapter1 = Adapter(
            name="test1",
            class_name="Test1",
            category="category1",
            installed=True,
        )
        adapter2 = Adapter(
            name="test2",
            class_name="Test2",
            category="category2",
            installed=False,
        )
        adapter_registry.set([adapter1, adapter2])

        assert get_installed_adapter("category1") == adapter1
        assert get_installed_adapter("category2") is None
        assert get_installed_adapter("non_existent") is None

    def test_get_installed_adapters(self) -> None:
        adapter1 = Adapter(
            name="test1",
            class_name="Test1",
            category="category1",
            installed=True,
        )
        adapter2 = Adapter(
            name="test2",
            class_name="Test2",
            category="category2",
            installed=False,
        )
        adapter3 = Adapter(
            name="test3",
            class_name="Test3",
            category="category3",
            installed=True,
        )
        adapter_registry.set([adapter1, adapter2, adapter3])

        installed_adapters = get_installed_adapters()
        assert len(installed_adapters) == 2
        assert adapter1 in installed_adapters
        assert adapter2 not in installed_adapters
        assert adapter3 in installed_adapters


class TestImportAdapter:
    def test_import_adapter_with_valid_category(self) -> None:
        mock_adapter_class = Mock()

        def mock_import_adapter_func(category: t.Optional[str] = None) -> Mock:
            return mock_adapter_class

        with patch("acb.adapters.import_adapter", side_effect=mock_import_adapter_func):
            from acb.adapters import import_adapter as patched_import_adapter

            result = patched_import_adapter("test_category")

            assert result == mock_adapter_class

    def test_import_adapter_with_invalid_category(self) -> None:
        def mock_import_adapter_func(category: t.Optional[str] = None) -> t.NoReturn:
            raise AdapterNotFound(f"{category} adapter not found")

        with patch("acb.adapters.import_adapter", side_effect=mock_import_adapter_func):
            from acb.adapters import import_adapter as patched_import_adapter

            with pytest.raises(AdapterNotFound):
                patched_import_adapter("invalid_category")

    def test_adapter_auto_detection_from_context(self) -> None:
        mock_adapter_class = Mock()

        def mock_import_adapter_func(category: t.Optional[str] = None) -> Mock:
            return mock_adapter_class

        with patch("acb.adapters.import_adapter", side_effect=mock_import_adapter_func):
            from acb.adapters import import_adapter as patched_import_adapter

            result = patched_import_adapter()

            assert result == mock_adapter_class

    def test_import_adapter_with_non_existent_module(self) -> None:
        mock_adapter = Mock()
        mock_adapter.name = "test"
        mock_adapter.class_name = "TestAdapter"
        mock_adapter.category = "test_category"
        mock_adapter.module = "acb.adapters.test_category.test"
        mock_adapter.path = AsyncPath(Path(__file__))
        mock_adapter.installed = False
        mock_adapter.enabled = True

        def mock_import_adapter_func(category: t.Optional[str] = None) -> t.NoReturn:
            raise AdapterNotFound(f"{category} adapter not found")

        with patch("acb.adapters.import_adapter", side_effect=mock_import_adapter_func):
            from acb.adapters import import_adapter as patched_import_adapter

            with pytest.raises(AdapterNotFound):
                patched_import_adapter("test_category")


class TestPathAdapters:
    def test_path_adapters_with_directory(self, tmp_path: Path) -> None:
        adapter_dir = tmp_path / "adapters"
        adapter_dir.mkdir()

        test_adapter_dir = adapter_dir / "test_adapter"
        test_adapter_dir.mkdir()

        (test_adapter_dir / "module1.py").touch()
        (test_adapter_dir / "module2.py").touch()
        (test_adapter_dir / "_private.py").touch()

        other_adapter_dir = adapter_dir / "other_adapter"
        other_adapter_dir.mkdir()
        (other_adapter_dir / "module3.py").touch()

        (adapter_dir / "__pycache__").mkdir()

        result = path_adapters(adapter_dir)

        assert "test_adapter" in result
        assert "other_adapter" in result
        assert "__pycache__" not in result

        assert len(result["test_adapter"]) == 2
        assert len(result["other_adapter"]) == 1

        test_adapter_modules = [m.name for m in result["test_adapter"]]
        assert "module1.py" in test_adapter_modules
        assert "module2.py" in test_adapter_modules
        assert "_private.py" not in test_adapter_modules


class TestRegisterAdapters:
    def test_simple(self) -> None:
        assert True

    def test_register_adapters_mock(self) -> None:
        with patch("acb.adapters.register_adapters") as mock_register:
            adapter1 = Adapter(
                name="adapter1",
                class_name="Adapter1",
                category="category1",
                module="acb.adapters.category1.adapter1",
                path=AsyncPath(Path(__file__)),
                pkg="acb",
                enabled=True,
            )

            adapter2 = Adapter(
                name="adapter2",
                class_name="Adapter2",
                category="category2",
                module="acb.adapters.category2.adapter2",
                path=AsyncPath(Path(__file__)),
                pkg="acb",
                enabled=False,
            )

            mock_register.return_value = [adapter1, adapter2]

            result = mock_register()

            assert len(result) == 2
            assert result[0].name == "adapter1"
            assert result[1].name == "adapter2"
            assert result[0].enabled
            assert not result[1].enabled



import asyncio
import os
import sys
import typing as t
from contextvars import ContextVar
from importlib import import_module, util
from inspect import currentframe, stack
from pathlib import Path

from aiopath import AsyncPath
from inflection import camelize
from msgspec.yaml import decode as yaml_decode
from pydantic import BaseModel
from acb.actions.encode import yaml_encode
from acb.depends import depends

root_path: AsyncPath = AsyncPath(Path.cwd())
tmp_path: AsyncPath = root_path / "tmp"


class Adapter(BaseModel, arbitrary_types_allowed=True):
    name: str
    class_name: str
    category: str
    pkg: str = "acb"
    module: str = ""
    enabled: bool = False
    installed: bool = False
    path: AsyncPath = AsyncPath(Path(__file__) / "adapters")

    def __str__(self) -> str:
        return self.__repr__()

    def __hash__(self) -> int:
        return hash((self.name, self.class_name, self.category, self.pkg, self.module))

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Adapter):
            return False
        return (
            self.name == other.name
            and self.class_name == other.class_name
            and self.category == other.category
            and self.pkg == other.pkg
            and self.module == other.module
        )


@t.runtime_checkable
class AdapterProtocol(t.Protocol):
    async def init(self) -> None: ...


adapter_registry: ContextVar[list[Adapter]] = ContextVar("adapter_registry", default=[])
_install_lock: ContextVar[list[str]] = ContextVar("install_lock", default=[])
_deployed: bool = os.getenv("DEPLOYED", "False").lower() == "true"
_testing: bool = os.getenv("TESTING", "False").lower() == "true"
settings_path: Path = Path(root_path / "settings")
app_settings_path: Path = settings_path / "app.yml"
adapter_settings_path: Path = settings_path / "adapters.yml"


class AdapterNotFound(Exception): ...


class AdapterNotInstalled(Exception): ...


def get_adapter(category: str) -> Adapter | None:
    adapter = next(
        (a for a in adapter_registry.get() if a.category == category and a.enabled),
        None,
    )
    return adapter


def get_adapters() -> list[Adapter]:
    return [a for a in adapter_registry.get() if a.enabled]


def get_installed_adapter(category: str) -> Adapter | None:
    adapter = next(
        (a for a in adapter_registry.get() if a.category == category and a.installed),
        None,
    )
    return adapter


def get_installed_adapters() -> list[Adapter]:
    return [a for a in adapter_registry.get() if a.installed]


async def _import_adapter(
    adapter_category: str, config: AdapterProtocol
) -> type[AdapterProtocol] | None:
    try:
        adapter = get_adapter(adapter_category)
        if adapter is None:
            raise AdapterNotFound(
                f"{adapter_category} adapter not found – check adapters.yml"
            )
    except AttributeError:
        raise AdapterNotFound(
            f"{adapter_category} adapter not found – check adapters.yml"
        )
    try:
        module = import_module(adapter.module)
    except ModuleNotFoundError:
        spec = util.spec_from_file_location(adapter.path.stem, adapter.path)
        module = util.module_from_spec(spec)  # type: ignore
        spec.loader.exec_module(module)
        sys.modules[adapter.name] = module

    adapter_class: type[AdapterProtocol] = getattr(module, adapter.class_name)

    if not adapter.installed and adapter.category not in _install_lock.get():
        _install_lock.get().append(adapter.category)
        try:
            adapter_settings_class_name = f"{adapter.class_name}Settings"
            adapter_settings_class = getattr(module, adapter_settings_class_name)
            adapter_settings = adapter_settings_class()
            setattr(config, adapter_category, adapter_settings)
            await depends.get(adapter_class).init()
        except Exception as e:
            raise AdapterNotInstalled(
                f"Failed to install {adapter.class_name} adapter: {e}"
            )
        else:
            adapter.installed = True
            if adapter_category != "logger":
                logger = depends.get()
                logger.info(f"{adapter.class_name} adapter installed")
        finally:
            _install_lock.get().remove(adapter.category)

    return adapter_class


def import_adapter(
    adapter_categories: t.Optional[str | list[str]] = None,
) -> (
    AdapterProtocol
    | list[AdapterProtocol | Exception | str]
    | str
    | Exception
    | list[str | Exception]
):
    config = depends.get()

    if isinstance(adapter_categories, str):
        adapter_categories = [adapter_categories]
    if not adapter_categories:
        try:
            adapter_categories = [
                c.strip()
                for c in (
                    stack()[1][4][0].split("=")[0].strip().lower()  # type: ignore
                ).split(",")
            ]
        except (IndexError, AttributeError, TypeError):
            raise ValueError(
                "Could not determine adapter categories from calling context"
            )
    try:
        imports: list[t.Coroutine[t.Any, t.Any, t.Type[AdapterProtocol] | None]] = [
            _import_adapter(c, config) for c in adapter_categories
        ]
        # classes = asyncio.run(asyncio.gather(*imports, return_exceptions=True))  # type: ignore
        classes = asyncio.run(asyncio.gather(*imports))  # type: ignore
        if isinstance(classes, str | Exception):
            classes = [classes]
        return classes[0] if len(adapter_categories) < 2 else classes
    except Exception as err:
        raise AdapterNotInstalled(err)


def path_adapters(path: Path) -> dict[str, list[Path]]:
    return {
        a.stem: [
            m for m in a.iterdir() if not m.name.startswith("_") and m.suffix == ".py"
        ]
        for a in path.iterdir()
        if a.is_dir() and not a.name.startswith("__")
    }


def create_adapter(path: Path) -> Adapter:
    return Adapter(
        name=path.stem,
        class_name=camelize(path.parent.stem),
        category=path.parent.stem,
        module=".".join(path.parts[-4:]).removesuffix(".py"),
        pkg=path.parent.parent.parent.stem,
        path=AsyncPath(path),
    )


def extract_adapter_modules(modules: list[Path], adapter: str) -> list[Adapter]:
    adapter_modules = [create_adapter(p) for p in modules]
    return [a for a in adapter_modules if a.category == adapter]


def register_adapters() -> list[Adapter]:
    adapters_path = (
        Path(currentframe().f_back.f_back.f_code.co_filename).parent / "adapters"
    )
    pkg_adapters = path_adapters(adapters_path)
    adapters: list[Adapter] = []
    for adapter_name, modules in pkg_adapters.items():
        _modules = extract_adapter_modules(modules, adapter_name)
        adapters.extend(_modules)
    if (
        not adapter_settings_path.exists()
        and not _deployed
        and not _testing
        and root_path.stem != "acb"
    ):
        settings_path.mkdir(exist_ok=True)
        categories = set()
        categories.update(a.category for a in adapter_registry.get() + adapters)
        adapter_settings_path.write_bytes(
            yaml_encode(
                {cat: None for cat in categories},
                sort_keys=True,
            )
        )
    _adapters = (
        {}
        if _testing or root_path.stem == "acb"
        else yaml_decode(adapter_settings_path.read_text())
    )
    _adapters.update(
        {a.category: None for a in adapters if a.category not in _adapters}
    )
    if not _testing and not root_path.stem == "acb":
        adapter_settings_path.write_bytes(yaml_encode(_adapters, sort_keys=True))
    enabled_adapters = {a: m for a, m in _adapters.items() if m}
    for a in [a for a in adapters if enabled_adapters.get(a.category) == a.name]:
        a.enabled = True
    registry = adapter_registry.get()
    for a in adapters:
        module = ".".join(a.module.split(".")[-2:])
        remove = next(
            (
                _a
                for _a in registry
                if ".".join(_a.module.split(".")[-2:]) == module and _a.pkg == "acb"
            ),
            None,
        )
        if remove:
            registry.remove(remove)
        if module in [".".join(_a.module.split(".")[-2:]) for _a in registry]:
            continue
        registry.append(a)
    return adapters



from pathlib import Path
from unittest.mock import Mock, patch

import pytest
from aiopath import AsyncPath
from acb.actions import (
    Action,
    ActionNotFound,
    ActionProtocol,
    Actions,
    action_registry,
    actions,
    create_action,
    register_actions,
)


class TestAction:
    def test_action_initialization(self) -> None:
        action = Action(name="test_action")

        assert action.name == "test_action"
        assert action.pkg == "acb"
        assert action.module == ""
        assert not action.methods
        assert isinstance(action.path, AsyncPath)

        custom_path = AsyncPath("/custom/path")
        action = Action(
            name="custom_action",
            pkg="custom_pkg",
            module="custom.module",
            methods=["method1", "method2"],
            path=custom_path,
        )

        assert action.name == "custom_action"
        assert action.pkg == "custom_pkg"
        assert action.module == "custom.module"
        assert action.methods == ["method1", "method2"]
        assert action.path == custom_path


class TestCreateAction:
    def test_create_action(self) -> None:
        test_path = Path("/test/pkg/actions/test_action")

        action = create_action(test_path)

        assert action.name == "test_action"
        assert action.pkg == "test"
        assert action.module == "pkg.actions.test_action"
        assert isinstance(action.path, AsyncPath)
        assert action.path == AsyncPath(test_path)


class TestActions:
    def test_getattr_success(self) -> None:
        test_actions = Actions()

        test_action = Action(name="existing_action")
        test_actions.__dict__["existing_action"] = test_action

        result = test_actions.existing_action

        assert result == test_action

    def test_getattr_failure(self) -> None:
        test_actions = Actions()

        with pytest.raises(ActionNotFound) as exc_info:
            test_actions.non_existent_action

        assert "Action non_existent_action not found" in str(exc_info.value)


class TestActionRegistry:
    def test_action_registry_context_var(self) -> None:
        registry = action_registry.get()

        assert isinstance(registry, list)

        original_length = len(registry)

        test_action = Action(name="test_action")
        token = action_registry.set([test_action])

        new_registry = action_registry.get()
        assert len(new_registry) == 1
        assert new_registry[0] == test_action

        action_registry.reset(token)

        reset_registry = action_registry.get()
        assert isinstance(reset_registry, list)
        assert len(reset_registry) == original_length


class TestActionProtocol:
    def test_action_protocol_compliance(self) -> None:
        assert isinstance(actions, ActionProtocol)

        class NonCompliant:
            pass

        non_compliant = NonCompliant()
        assert not isinstance(non_compliant, ActionProtocol)

        class Compliant:
            def __getattr__(self, item: str) -> Action:
                return Action(name=item)

        compliant = Compliant()
        assert isinstance(compliant, ActionProtocol)


class TestRegisterActions:
    def test_register_actions(self) -> None:
        mock_caller_file = "/test/pkg/main.py"
        mock_actions_path = Path("/test/pkg/actions")
        mock_action_dir = Path("/test/pkg/actions/test_action")

        mock_dirs = {"test_action": mock_action_dir}

        mock_module = Mock()
        mock_module.__all__ = ["test_method"]

        test_method = Mock()
        mock_module.test_method = test_method

        mock_registry_list = []

        with (
            patch("acb.actions.currentframe") as mock_frame,
            patch("acb.actions.Path") as mock_path_cls,
            patch("acb.actions.action_registry") as mock_registry,
            patch("acb.actions.import_module") as mock_import,
            patch("acb.actions.create_action") as mock_create_action,
            patch.object(Path, "exists") as mock_exists,
            patch.object(Path, "iterdir") as mock_iterdir,
            patch.object(Path, "is_dir") as mock_is_dir,
        ):
            mock_frame.return_value.f_back.f_back.f_code.co_filename = mock_caller_file
            mock_path_cls.return_value = mock_actions_path
            mock_exists.return_value = True

            mock_iterdir.return_value = [
                Path(f"{mock_actions_path}/{name}") for name in mock_dirs.keys()
            ]

            mock_is_dir.return_value = True

            mock_registry.get.return_value = mock_registry_list

            test_action = Action(
                name="test_action",
                module="pkg.actions.test_action",
                pkg="pkg",
                path=AsyncPath(mock_action_dir),
            )
            mock_create_action.return_value = test_action

            mock_import.return_value = mock_module

            result = register_actions()

            assert len(result) == 1
            assert result[0] == test_action

            mock_registry.get.assert_called_once()
            assert len(mock_registry_list) == 1
            assert mock_registry_list[0] == test_action

            mock_import.assert_called_once_with(test_action.module)

            assert hasattr(actions, "test_method")

    def test_register_actions_with_missing_module(self) -> None:
        mock_caller_file = "/test/pkg/main.py"
        mock_actions_path = Path("/test/pkg/actions")
        mock_action_dir = Path("/test/pkg/actions/test_action")

        mock_dirs = {"test_action": mock_action_dir}

        mock_spec = Mock()
        mock_loader = Mock()
        mock_spec.loader = mock_loader

        mock_module = Mock()
        mock_module.__all__ = ["test_method"]

        test_method = Mock()
        mock_module.test_method = test_method

        mock_registry_list = []

        with (
            patch("acb.actions.currentframe") as mock_frame,
            patch("acb.actions.Path") as mock_path_cls,
            patch("acb.actions.action_registry") as mock_registry,
            patch("acb.actions.import_module") as mock_import,
            patch("acb.actions.util.spec_from_file_location") as mock_spec_from_file,
            patch("acb.actions.util.module_from_spec") as mock_module_from_spec,
            patch("acb.actions.create_action") as mock_create_action,
            patch.object(Path, "exists") as mock_exists,
            patch.object(Path, "iterdir") as mock_iterdir,
            patch.object(Path, "is_dir") as mock_is_dir,
        ):
            mock_frame.return_value.f_back.f_back.f_code.co_filename = mock_caller_file
            mock_path_cls.return_value = mock_actions_path
            mock_exists.return_value = True

            mock_iterdir.return_value = [
                Path(f"{mock_actions_path}/{name}") for name in mock_dirs.keys()
            ]

            mock_is_dir.return_value = True

            mock_registry.get.return_value = mock_registry_list

            test_action = Action(
                name="test_action",
                module="pkg.actions.test_action",
                pkg="pkg",
                path=AsyncPath(mock_action_dir),
            )
            mock_create_action.return_value = test_action

            mock_import.side_effect = ModuleNotFoundError("Module not found")

            mock_spec_from_file.return_value = mock_spec

            mock_module_from_spec.return_value = mock_module

            result = register_actions()

            assert len(result) == 1
            assert result[0] == test_action

            mock_registry.get.assert_called_once()
            assert len(mock_registry_list) == 1
            assert mock_registry_list[0] == test_action

            mock_spec_from_file.assert_called_once_with(
                test_action.path.stem, test_action.path
            )
            mock_module_from_spec.assert_called_once_with(mock_spec)
            mock_loader.exec_module.assert_called_once_with(mock_module)

            assert hasattr(actions, "test_method")



import typing as t

from acb.config import AdapterBase, Settings


class ModelsBaseSettings(Settings): ...


class ModelsProtocol(t.Protocol):
    sql: t.Any = ...
    nosql: t.Any = ...

    class SqlModels: ...

    class NosqlModels: ...


class ModelsBase(AdapterBase):
    class SqlModels: ...

    sql = SqlModels()

    class NosqlModels: ...

    nosql = NosqlModels()



from acb.config import AdapterBase, import_adapter
from acb.config import AppSettings as AppConfigSettings

Logger = import_adapter()


class AppBaseSettings(AppConfigSettings): ...


class AppBase(AdapterBase): ...



import asyncio
import logging
from contextlib import suppress
from functools import wraps
from pathlib import Path
from time import perf_counter
from typing import Any, Callable, Protocol, TypeVar, cast

from aioconsole import aprint
from devtools import pformat
from icecream import colorize, supportTerminalColorsInWindows
from icecream import ic as debug
from acb.adapters import import_adapter
from acb.config import adapter_registry
from acb.depends import depends

__all__ = [
    "get_calling_module",
    "patch_record",
    "colorized_stderr_print",
    "print_debug_info",
    "timeit",
    "debug",
]

Logger = import_adapter()

config = depends.get()


def get_calling_module() -> Path | None:
    try:
        mod = logging.currentframe().f_back.f_back.f_back.f_code.co_filename
        mod = Path(mod).parent
        debug_mod = getattr(config.debug, mod.stem, None)
        return mod if debug_mod else None
    except (AttributeError, TypeError):
        return None


@depends.inject
def patch_record(
    mod: Path,
    msg: str,
    logger: Logger = depends(),
) -> None:
    with suppress(Exception):
        has_loguru = any(
            a.category == "logger" and a.name == "loguru"
            for a in adapter_registry.get()
        )
        if has_loguru:
            logger.patch(lambda record: record.update(name=mod.name)).debug(msg)


def colorized_stderr_print(s: str) -> None:
    colored = colorize(s)
    with supportTerminalColorsInWindows():
        asyncio.run(aprint(colored, use_stderr=True))


def print_debug_info(msg: str) -> Any:
    mod = get_calling_module()
    if mod:
        if config.deployed or config.debug.production:
            patch_record(mod, msg)
        else:
            colorized_stderr_print(msg)


debug_args = dict(
    outputFunction=print_debug_info,
    argToStringFunction=lambda o: pformat(o, highlight=False),
)
debug.configureOutput(
    prefix="    debug:  ",
    includeContext=True,
    **debug_args,
)
if config.deployed or config.debug.production:
    debug.configureOutput(
        prefix="",
        includeContext=False,
        **debug_args,
    )


class TimedFunction(Protocol):
    def __call__(self, *args: Any, **kwargs: Any) -> Any: ...


T = TypeVar("T", bound=Callable[..., Any])


@depends.inject
def timeit(func: T, logger: Logger = depends()) -> T:
    @wraps(func)
    def wrapped(*args: Any, **kwargs: Any) -> Any:
        start = perf_counter()
        result = func(*args, **kwargs)
        end = perf_counter()
        logger.debug(f"Function '{func.__name__}' executed in {end - start} s")
        return result

    return cast(T, wrapped)



from pydantic import field_validator
from acb.config import AdapterBase, Settings


class MonitoringBaseSettings(Settings):
    traces_sample_rate: float = 0

    @field_validator("traces_sample_rate")
    @classmethod
    def check_traces_sample_rate(cls, v: float) -> float:  # noqa: F841
        if v > 1 or v < 0:
            raise ValueError("sample rate must be between 0 and 1")
        return v


class MonitoringBase(AdapterBase): ...
